// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*



////////////////////////////////////////////////////////////////////////


#ifndef  __VERIF_CONTINGENCY_TABLE_H__
#define  __VERIF_CONTINGENCY_TABLE_H__


////////////////////////////////////////////////////////////////////////


#include <iostream>

#include "vx_math/vx_math.h"


////////////////////////////////////////////////////////////////////////


class TTContingencyTable;   //  forward reference


////////////////////////////////////////////////////////////////////////


   //
   //  general contingency table
   //


class ContingencyTable {

   protected:

      void init_from_scratch();

      void assign(const ContingencyTable &);

      int rc_to_n(int r, int c) const;

      int Nrows;
      int Ncols;

      int * E;

      char * Name;

   public:

      ContingencyTable();
      virtual ~ContingencyTable();
      ContingencyTable(const ContingencyTable &);
      ContingencyTable & operator=(const ContingencyTable &);

      void clear();

      void zero_out();

      virtual void dump(ostream & out, int depth = 0) const;

         //
         //  condition on an event
         //

      TTContingencyTable condition_on(int) const;

         //
         //  set attributes
         //

      virtual void set_size(int);
      virtual void set_size(int NR, int NC);

      void set_name(const char *);

         //
         //  get attributes
         //

      int nrows() const;
      int ncols() const;

      const char * name() const;

         //
         //  set counts
         //

      void set_entry(int row, int col, int value);

         //
         //  increment counts
         //

      void inc_entry(int row, int col);

         //
         //  get counts
         //

      int total() const;

      int row_total(int row) const;
      int col_total(int col) const;

      int entry(int row, int col) const;

      int  largest_entry() const;
      int smallest_entry() const;

};


////////////////////////////////////////////////////////////////////////


inline int ContingencyTable::nrows() const { return ( Nrows ); }
inline int ContingencyTable::ncols() const { return ( Ncols ); }

inline const char * ContingencyTable::name() const { return ( Name ); }


////////////////////////////////////////////////////////////////////////


   //
   //  2 x 2 contingency table
   //


class TTContingencyTable : public ContingencyTable {

   public:

      TTContingencyTable();
      virtual ~TTContingencyTable();
      TTContingencyTable(const TTContingencyTable &);
      TTContingencyTable & operator=(const TTContingencyTable &);

      void set_size(int);
      void set_size(int NR, int NC);

         //
         //  set counts
         //

      void set_fn_on(int);
      void set_fy_on(int);

      void set_fn_oy(int);
      void set_fy_oy(int);

         //
         //  increment counts
         //

      void inc_fn_on();
      void inc_fy_on();

      void inc_fn_oy();
      void inc_fy_oy();

         //
         //  get counts
         //

      int fy_oy() const;
      int fy_on() const;
      int fn_oy() const;
      int fn_on() const;

      int fy() const;
      int fn() const;
      
      int oy() const;
      int on() const;

      int n() const;
      
         //
         //  FHO rates where:
         //  f_rate = FY/N
         //  h_rate = fy_oy/N
         //  o_rate = OY/N
         //
      
      double f_rate   (int &) const;
      double h_rate   (int &) const;
      double o_rate   (int &) const;
      
         //
         //  Raw counts as proportions of the
         //  total count.
         //

      double fy_oy_tp  (int &) const;
      double fy_on_tp  (int &) const;
      double fn_oy_tp  (int &) const;
      double fn_on_tp  (int &) const;
      
      double fy_tp     (int &) const;
      double fn_tp     (int &) const;
      
      double oy_tp     (int &) const;
      double on_tp     (int &) const;      
      
         //
         //  Raw counts as proportions of the
         //  total forecast yes count.
         //

      double fy_oy_fp (int &) const;
      double fy_on_fp (int &) const;
      double fn_oy_fp (int &) const;
      double fn_on_fp (int &) const;

         //
         //  Raw counts as proportions of the
         //  total observation yes count.
         //

      double fy_oy_op (int &) const;
      double fy_on_op (int &) const;
      double fn_oy_op (int &) const;
      double fn_on_op (int &) const;
      
         //
         //  Contingency Table Statistics with confidence intervals
         //  when applicable.
         //
      
      double baser      (int &) const;
      double baser_ci   (int &, double alpha, double &cl, double &cu) const;
      double fmean      (int &) const;
      double fmean_ci   (int &, double alpha, double &cl, double &cu) const;
      double accuracy   (int &) const;
      double accuracy_ci(int &, double alpha, double &cl, double &cu) const;
      double fbias      (int &) const;
      double pod_yes    (int &) const;
      double pod_yes_ci (int &, double alpha, double &cl, double &cu) const;
      double pod_no     (int &) const;
      double pod_no_ci  (int &, double alpha, double &cl, double &cu) const;
      double pofd       (int &) const;
      double pofd_ci    (int &, double alpha, double &cl, double &cu) const;
      double far        (int &) const;
      double far_ci     (int &, double alpha, double &cl, double &cu) const;
      double csi        (int &) const;
      double csi_ci     (int &, double alpha, double &cl, double &cu) const;
      double gss        (int &) const;
      double hk         (int &) const;
      double hk_ci      (int &, double alpha, double &cl, double &cu) const;
      double hss        (int &) const;
      double odds       (int &) const;
      double odds_ci    (int &, double alpha, double &cl, double &cu) const;
};


////////////////////////////////////////////////////////////////////////

extern TTContingencyTable finley();

extern TTContingencyTable finley_always_no();

////////////////////////////////////////////////////////////////////////


   //
   //  this is the layout on page 239 of
   //
   //   "Statistical Methods in the Atmospheric Sciences" (1st ed)
   //
   //     by Daniel S. Wilks
   //


static const int OY_col = 0;
static const int ON_col = 1;

static const int FY_row = 0;
static const int FN_row = 1;


////////////////////////////////////////////////////////////////////////


#endif   //  __VERIF_CONTINGENCY_TABLE_H__


////////////////////////////////////////////////////////////////////////


