// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <regex.h>

#include "vx_met_util/constants.h"
#include "vx_met_util/apply_mask.h"

///////////////////////////////////////////////////////////////////////////////

void parse_grid_mask(const char *mask_grid_str, const Grid &grid,
                     WrfData &mask_wd, char *&mask_name) {
   int x, y;
   Grid mask_grid;

   // Initialize the WrfData masking object
   mask_wd.set_size(grid.nx(), grid.ny());
   for(x=0; x<grid.nx(); x++) {
      for(y=0; y<grid.ny(); y++) {
         mask_wd.put_xy_int(wrfdata_int_data_max, x, y);
      }
   }

   // Store the name of the masking grid
   mask_name = new char [strlen(mask_grid_str)+1];
   strcpy(mask_name, mask_grid_str);

   //
   // Search for the special full domain string
   //
   if(strcmp(full_domain_str, mask_grid_str) == 0) {
      mask_grid = grid;
   }
   //
   // Otherwise, search for the predefined grids
   //
   else if(find_grid(mask_name, mask_grid) != 0) {
      cerr << "\n\nERROR: parse_grid_mask() -> "
           << "the mask_grid requested \"" << mask_grid_str
           << "\" is not defined.\n\n" << flush;
      exit(1);
   }

   apply_grid_mask(&grid, &mask_grid, mask_wd, 0);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void parse_poly_mask(const char *mask_poly_str, const Grid &grid,
                     WrfData &mask_wd, char *&mask_name) {
   int x, y;
   char mask_poly_tmp[max_str_len];
   Polyline mask_poly;

   // Replace any instances of MET_BASE with it's expanded value
   replace_string(met_base_str, MET_BASE, mask_poly_str, mask_poly_tmp);

   // Initialize the WrfData masking object
   mask_wd.set_size(grid.nx(), grid.ny());
   for(x=0; x<grid.nx(); x++) {
      for(y=0; y<grid.ny(); y++) {
         mask_wd.put_xy_int(wrfdata_int_data_max, x, y);
      }
   }

   // Initialize the masking polyline
   mask_poly.clear();

   // Parse out the polyline from the file
   parse_latlon_poly_file(mask_poly_tmp, mask_poly);

   // Store the name of the masking polyline
   mask_name = new char [strlen(mask_poly.name)+1];
   strcpy(mask_name, mask_poly.name);

   // Apply the polyline mask to the WrfData object
   apply_poly_mask_latlon(&mask_poly, &grid, mask_wd, 0);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void apply_grid_mask(const Grid *grid, const Grid *mask_grid, WrfData &wd, int v_int) {
   int x, y;
   double lat, lon, mask_x, mask_y;
   int in_count, out_count;

   //
   // If the grid pointer is NULL or the grid's dimensions are zero,
   // no masking region is defined and there is nothing to do
   //
   if(!mask_grid || mask_grid->nx() == 0 || mask_grid->ny() == 0) {
      cout << "***WARNING*** void apply_grid_mask() -> "
           << "no grid masking applied since the masking grid "
           << "specified is empty\n"
           << flush;

      return;
   }

   //
   // For each point of the original grid, convert it to (lat, lon) and then
   // to an (x, y) point on the mask grid. If the (x, y) is on within the
   // bounds of the mask grid, retain it's value.  Otherwise, replace it
   // with the value of v_int
   //
   in_count = out_count = 0;
   for(x=0; x<wd.get_nx(); x++) {
      for(y=0; y<wd.get_ny(); y++) {

         grid->xy_to_latlon(x, y, lat, lon);
         mask_grid->latlon_to_xy(lat, lon, mask_x, mask_y);

         if(mask_x < 0 || mask_x >= mask_grid->nx() ||
            mask_y < 0 || mask_y >= mask_grid->ny()) {
            out_count++;
            wd.put_xy_int(v_int, x, y);
         }
         else {
            in_count++;
         }
      } // for y
   } // for x

   return;
}

///////////////////////////////////////////////////////////////////////////////

void apply_poly_mask_xy(const Polyline *p, WrfData &wd, int v_int) {
   int x, y;
   int in_count, out_count;

   //
   // If the polyline pointer is NULL or the polyline contains less than 3
   // points, no masking region is defined and there is nothing to do
   //
   if(!p || p->n_points < 3) {
      cout << "***WARNING*** void apply_poly_mask_xy() -> "
           << "no polyline masking applied since the masking polyline "
           << "contains fewer than 3 points\n"
           << flush;

      return;
   }

   //
   // Replace any grid points not inside the masking polygon with the value
   // of v_int
   //
   in_count = out_count = 0;
   for(x=0; x<wd.get_nx(); x++) {
      for(y=0; y<wd.get_ny(); y++) {

         if(!(p->is_inside(x, y))) {
            out_count++;
            wd.put_xy_int(v_int, x, y);
         }
         else {
            in_count++;
         }
      } // for y
   } // for x

   return;
}

///////////////////////////////////////////////////////////////////////////////

void apply_poly_mask_latlon(const Polyline *p, const Grid *grid, WrfData &wd, int v_int) {
   int x, y;
   double lat, lon;
   int in_count, out_count;

   //
   // If the polyline pointer is NULL or the polyline contains less than 3
   // points, no masking region is defined and there is nothing to do
   //
   if(!p || p->n_points < 3) {
      cout << "***WARNING*** void apply_poly_mask_latlon() -> "
           << "no polyline masking applied since the masking polyline "
           << "contains fewer than 3 points\n"
           << flush;

      return;
   }

   //
   // Replace any grid points who's corresponding lat/lon coordinates are
   // not inside the masking lat/lon polygon with the value of v_int
   //
   in_count = out_count = 0;
   for(x=0; x<wd.get_nx(); x++) {
      for(y=0; y<wd.get_ny(); y++) {

         grid->xy_to_latlon(x, y, lat, lon);

         if(!(p->is_inside(lon, lat))) {
            out_count++;
            wd.put_xy_int(v_int, x, y);
         }
         else {
            in_count++;
         }
      } // for y
   } // for x

   return;
}

///////////////////////////////////////////////////////////////////////////////
